// src/backend/sgpemv2/resource_policy_priority.hh - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#ifndef RESOURCE_POLICY_PRIORITY_HH
#define RESOURCE_POLICY_PRIORITY_HH 1

#include <sgpemv2/resource_policy.hh>

namespace sgpem
{

  /// \brief A resource allocation policy which provides a better service
  /// to higer priority threads.
  ///
  /// A resource allocation policy which provides a better service
  /// to higer priority threads. This policy sorts the request queues
  /// following the priority of the owners of the requests.
  class ResourcePolicyPriority;

  class SG_DLLEXPORT ResourcePolicyPriority : public ResourcePolicy
  {
  public:

    /// \brief Standard virtual destructor.
    ///
    /// Standard virtual destructor.
    virtual ~ResourcePolicyPriority();

    /// \brief Initializes the inner components of the policy.
    ///
    /// Does nothing.
    virtual void configure() throw(UserInterruptException);

    /// \brief Sorts the subrequest queue letting higher priority threads to be the first ones
    /// to get the resources.
    ///
    /// Sorts the subrequest queue letting higher priority threads to be the first ones
    /// to get the resources.
    ///
    /// \param environment the environment on which the policy applies.
    /// \param queue the queue where a subrequest has just been added.
    /// \param sr the subrequest which has just been added.
    virtual void enforce(Environment& environment, Environment::SubRequestQueue& queue, SubRequest& sr) throw(UserInterruptException);

    /// \brief Returns a description of the policy.
    ///
    /// Returns a description of the policy.
    /// \return a description of the policy.
    virtual Glib::ustring get_description() const;

    /// \brief Returns the name of the policy.
    ///
    /// Returns the name of the policy.
    /// \return the name of the policy.
    virtual Glib::ustring get_name() const;


    /// \brief Activates the policy.
    ///
    /// Does nothing.
    virtual void activate();

    /// \brief Deactivates the policy.
    ///
    /// Does nothing.
    virtual void deactivate();

    /// \brief Returns the parameters used to customize the policy.
    ///
    /// Returns the parameters used to customize the policy.
    /// The user may modify them directly.
    /// \return the parameters used to customize the policy.
    PolicyParameters& get_parameters();

  protected:
    PolicyParameters _parameters;
  };

}//~ namespace sgpem


#endif
